from c4v.scraper.scraper    import bulk_scrape, scrape
from c4v.scraper.settings   import INSTALLED_CRAWLERS, SUPPORTED_DOMAINS, INSTALLED_SCRAPERS
from c4v.scraper.scraped_data_classes.scraped_data import ScrapedData



"""
    This file implements a class that manages local storage 
    using regular files.
"""
# Python
import sqlite3
from typing import Any, Iterator, Dict, List, Tuple
import dataclasses
import datetime

# Local imports
from c4v.scraper.persistency_manager.base_persistency_manager import (
    BasePersistencyManager,
)
from c4v.scraper.scraped_data_classes.scraped_data import Labels, ScrapedData, Sources
from c4v.config import settings

DATE_FORMAT = settings.date_format


class SqliteManager(BasePersistencyManager):
    """
        Use SQLite based local storage 
        to manage persistency.
    """

    def __init__(self, db_path: str):
        self._db_path = db_path
        self._migrate(db_path)

    def _migrate(self, db_path: str):
        """
            Perform a database migration, ensuring that tables are as specified in this class,
            use db specified by given path
        """
        with sqlite3.connect(db_path) as connection:
            cursor = connection.cursor()  # connect to cursor

            # All these migations are "if not exists" so no change is performed when they already exist
            self._create_scraped_data_table(cursor)  # create scraped data table
            self._create_category_table(
                cursor
            )  # category table (including many to many)
            connection.commit()  # save changes

    def _create_scraped_data_table(self, cursor: sqlite3.Cursor):
        """
            Create scraped data with the structure as in ScrapedData model, perform changes with 
            provided cursor
        """
        cursor.execute(
            "CREATE TABLE IF NOT EXISTS  scraped_data(\
            url PRIMARY KEY,\
            last_scraped DATETIME NULL,\
            title TEXT NULL,\
            content TEXT NULL,\
            author TEXT NULL,\
            date DATETIME NULL,\
            label TEXT NULL, \
            source TEXT NULL\
            );"
        )

    def _create_category_table(self, cursor: sqlite3.Cursor):
        """
            Create category table, including a many to many relationship between categories.
            Perform changes with provided cursor
        """
        cursor.execute(
            "CREATE TABLE IF NOT EXISTS category(\
            name PRIMARY KEY\
        );"
        )
        cursor.execute(
            "CREATE TABLE IF NOT EXISTS  category_to_data (\
                scraped_data_url,\
                category_name,\
                FOREIGN KEY (scraped_data_url) REFERENCES scraped_data(url) ON DELETE CASCADE,\
                FOREIGN KEY (category_name) REFERENCES category(name) ON DELETE CASCADE,\
                CONSTRAINT uniq UNIQUE (scraped_data_url, category_name)\
            );"
        )

    def get_all(
        self, limit: int = -1, scraped: bool = None, order_by: List[str] = None
    ) -> Iterator[ScrapedData]:

        # Retrieve all data stored
        with sqlite3.connect(self._db_path) as connection:
            cursor = connection.cursor()
            # Order
            if order_by:
                parsed_orders = " ORDER BY " + ", ".join(
                    [
                        f"{name} {order}"
                        for (order, name) in (
                            self._parse_order_and_field_from_order_by_str(order_by=ord)
                            for ord in order_by
                        )
                    ]
                )
            else:
                parsed_orders = ""

            # if scraped = true, take only scraped. If false, take non-scraped. If none, take all
            if scraped:
                new_cur = cursor.execute(
                    f"SELECT * FROM scraped_data WHERE last_scraped IS NOT NULL{parsed_orders};",
                )
            elif scraped == False:
                new_cur = cursor.execute(
                    f"SELECT * FROM scraped_data WHERE last_scraped IS NULL{parsed_orders};",
                )
            else:
                new_cur = cursor.execute(f"SELECT * FROM scraped_data{parsed_orders};",)

            # If limit less than 0, then take as much as you can
            if limit < 0:
                res = new_cur.fetchall()
            else:
                res = new_cur.fetchmany(limit)

            for row in res:
                # Decompose row
                (url, last_scraped, title, content, author, date, label, source) = row

                if label:
                    try:
                        label = Labels(label)
                    except:  # not a known label
                        label = Labels.UNKNOWN

                if source:
                    try:
                        source = Sources(source)
                    except:  # unknown source
                        source = Sources.UNKOWN

                # parse date to datetime:
                try:
                    last_scraped = (
                        datetime.datetime.strptime(last_scraped, DATE_FORMAT)
                        if last_scraped
                        else last_scraped
                    )
                except ValueError as _:  # In case it fails using a format not valid for python3.6
                    for i in range(len(last_scraped) - 1, -1, -1):
                        if last_scraped[i] == ":":
                            last_scraped = last_scraped[:i] + last_scraped[i + 1 :]
                            break
                    last_scraped = (
                        datetime.datetime.strptime(last_scraped, DATE_FORMAT)
                        if last_scraped
                        else last_scraped
                    )

                categories = [
                    category
                    for (category,) in cursor.execute(
                        "SELECT category_name FROM category_to_data WHERE scraped_data_url=?;",
                        [url],
                    )
                ]

                yield ScrapedData(
                    url=url,
                    last_scraped=last_scraped,
                    title=title,
                    content=content,
                    author=author,
                    date=date,
                    categories=categories,
                    label=label,
                    source=source,
                )

    def filter_scraped_urls(self, urls: List[str]) -> List[str]:

        # connect to db and check for each url if such url was scraped,
        # checking its last_scrape field

        res = []
        with sqlite3.connect(self._db_path) as connection:

            cursor = connection.cursor()
            for url in urls:

                # if last_scraped is null, then it wasn't scraped
                is_there = cursor.execute(
                    "SELECT 1 FROM scraped_data WHERE url=? AND last_scraped IS NOT NULL",
                    [url],
                ).fetchone()

                if not is_there:
                    res.append(url)

        return res

    def filter_known_urls(self, urls: List[str]) -> List[str]:

        # connect to db and check for each url if such url was scraped,
        # checking its last_scrape field

        res = []
        with sqlite3.connect(self._db_path) as connection:

            cursor = connection.cursor()
            for url in urls:

                # if last_scraped is null, then it wasn't scraped
                is_there = cursor.execute(
                    "SELECT 1 FROM scraped_data WHERE url=?", [url],
                ).fetchone()

                if not is_there:
                    res.append(url)

        return res

    def was_scraped(self, url: str) -> bool:

        # Connect to db and check if object was scraped
        with sqlite3.connect(self._db_path) as connection:
            cursor = connection.cursor()

            obj = cursor.execute(
                "SELECT ? FROM scraped_data WHERE last_scraped IS NOT NULL", [url]
            ).fetchone()

            return obj != None

    def save(self, url_data: List[ScrapedData]):
        # Save data into database

        # If nothing to do, just return
        if not url_data:
            return

        with sqlite3.connect(self._db_path) as connection:
            cursor = connection.cursor()

            # I think that this should be removed
            # for data in url_data:
            #     new_data = dataclasses.asdict(data)
            #     new_data["last_scraped"] = (
            #         datetime.datetime.strftime(data.last_scraped, DATE_FORMAT)
            #         if data.last_scraped
            #         else data.last_scraped
            #     )

            data_to_insert = [dataclasses.asdict(data) for data in url_data]
            for data in data_to_insert:
                label: Labels = data["label"]
                data["label"] = label.value if label else label
                source: Sources = data["source"]
                data["source"] = source.value if source else source

            cursor.executemany(
                "INSERT OR REPLACE INTO scraped_data VALUES (:url, :last_scraped, :title, :content, :author, :date, :label, :source)",
                data_to_insert,
            )

            for data in url_data:
                # insert new categories
                # may have none if category field is set to None
                if data.categories is None:
                    continue

                cursor.executemany(
                    "INSERT OR IGNORE INTO category VALUES (?)",
                    [(cat,) for cat in data.categories],
                )

                # insert many to many relationship
                cursor.executemany(
                    "INSERT OR IGNORE INTO category_to_data VALUES (?, ?)",
                    [(data.url, category) for category in data.categories],
                )

            # save changes
            connection.commit()

    def delete(self, urls: List[str]):

        # Bulk delete provided urls
        with sqlite3.connect(self._db_path) as connection:
            cursor = connection.cursor()
            cursor.executemany(
                "DELETE FROM scraped_data WHERE url=?", [(url,) for url in urls]
            )
            connection.commit()

    def _parse_order_and_field_from_order_by_str(
        self, order_by: str
    ) -> Tuple[str, str]:
        """
            Parse order and name of field from the given str
            Parameters:
                + order_by : str = A string with the following format:
                    (+|-)<field_name> where field_name is the name of a field in ScrapedData,
                    '+' corresponds to a an ascending order, and '-' corresponds to a descending order
            Return:
                Tuple such that
                ('ASC' | 'DESC', <field_name>) 
        """

        # Check valid format
        if not order_by:
            raise ValueError(
                "Invalid value for order_by string, I'm getting an empty string"
            )
        elif order_by[0] != "-" and order_by[0] != "+":
            raise ValueError(
                f"Invalid value for order_by string, first char should be the order, on of: [-, +]. Provided string: {order_by}"
            )
        elif len(order_by) < 2:
            raise ValueError(
                f"Invalid value for order_by string, name of field not provided. Given string: {order_by}"
            )

        # Parsing order
        order = "ASC" if order_by[0] == "+" else "DESC"

        # Parsing field name
        valid_fields = [d.name for d in dataclasses.fields(ScrapedData)]
        field_name = order_by[1:]

        # Raise error if not a valid string
        if field_name not in valid_fields:
            valid_fields_str = "\n".join([f"\t* {s}\n" for s in valid_fields])
            raise ValueError(
                f"Invalid value for order_by string, provided name doesn't match any field in ScrapedData. Provided field: {field_name}.\nAvailable fields:\n{valid_fields_str}"
            )

        return (order, field_name)


def _parse_dict_to_url_data(obj: Dict[str, Any]) -> ScrapedData:
    """
    Parse dict object into a new object instance
    """
    return ScrapedData(**obj)



"""
    The ScrapedData class is a target scheme for scraped data output, it describes how should stored in the db
"""

# Python imports
from dataclasses import dataclass, asdict, field

from typing import List, Dict, Any
from datetime import datetime
import json
import enum

# Local imports
from c4v.config import settings


class Labels(enum.Enum):
    """
        Every possible label
    """

    IRRELEVANTE: str = "IRRELEVANTE"
    DENUNCIA_FALTA_DEL_SERVICIO: str = "PROBLEMA DEL SERVICIO"
    UNKNOWN: str = "UNKNOWN"

    @classmethod
    def labels(cls) -> List[str]:
        """
            Get list of labels as strings
        """
        return [l.value for l in cls]


class Sources(enum.Enum):
    """
        Every possible source
    """

    UNKOWN: str = "UNKNOWN"
    SCRAPING: str = "SCRAPING"
    CLIENT: str = "CLIENT"


@dataclass()
class ScrapedData:
    """
        This is a general data format class, 
        every data format for other scrapers could have 
        additional fields according to its needs and
        scrapable data, but then they should be able to 
        convert themselves into this format, possibly leaving a 
        few fields as None. Thus, we can be able to 
        easily map from a scrapers's output to a database 
        scheme
    """

    url: str
    last_scraped: datetime = None
    title: str = None
    content: str = None
    author: str = None
    categories: List[str] = field(default_factory=list)
    date: str = None
    label: Labels = None
    source: Sources = None

    def pretty_repr(self, max_content_len: int = -1) -> str:
        """
            Return a human-readable representation of this data.
            Truncate content if requested
        """

        # create categories string
        categories = "".join(map(lambda s: f"\t+ {s}\n", self.categories))

        max_content_len = max_content_len if max_content_len > 0 else len(self.content)

        # create body content:
        content = self.content
        if max_content_len < len(self.content):
            content = content[:max_content_len] + "..."

        return f"title: {self.title}\nauthor: {self.author}\ndate: {self.date}\ncategories:\n{categories}\nlabel: {self.label}\nsource: {self.source}\ncontent:\n\t{content}"

    def __hash__(self) -> int:
        return (
            self.url,
            self.last_scraped,
            self.title,
            self.content,
            self.author,
            self.date,
            self.label,
            self.source,
        ).__hash__()


class ScrapedDataEncoder(json.JSONEncoder):
    """
        Encoder to turn this file into json format
    """

    def default(self, obj: ScrapedData) -> Dict[str, Any]:
        if isinstance(obj, ScrapedData):
            return asdict(obj)
        elif isinstance(obj, datetime):
            return datetime.strftime(obj, settings.date_format)

        return json.JSONEncoder.default(self, obj)



"""
    Use this script to process output data from the primicia_irrelevant_news_scraping.py script
    The file, in order to match the one used for trainning and provided by the client, should provide the following collumns
    * label
    * title
    * content
    * date
    * scraped_Date
    * author
    * url

    Expected arguments:
        csv_file_name : str = name of csv file with corresponding dataframe to use as basis
"""
import pandas as pd
import sys
import os
from pandas.core.frame import DataFrame

# Get command line argument
if len(sys.argv) < 2:
    print("Missing csv filename argument", file=sys.stderr)
    exit(1)

csv_file_name = sys.argv[1]  # name of csv file
try:
    with open(csv_file_name) as file:
        df: DataFrame = pd.read_csv(file)
except Exception as e:
    print(
        f"[Error] Could not write to provided file: {csv_file_name}.\n\tError: {e}",
        file=sys.stderr,
    )

# Remove empty content
print("Removing rows with empty values...")

df.dropna(inplace=True, axis=0)
df["content"].drop(
    df[(df.content.str.len() < 1) | (df.title.str.len() < 1)].index, inplace=True
)

# Add label column as irrelevant for every row
print("Adding irrelevant label to every row...")
df["label"] = [["IRRELEVANTE"]] * len(df)

# Remove extra linejumps
print("Removing extra line jumps from article bodies...")


def strip_extra_linejumps(s: str) -> str:
    return "\n".join([line for line in s.splitlines() if line.strip() != ""])


df["content"] = df["content"].map(strip_extra_linejumps)

# Remove duplicates if any
print("Removing duplicates rows if any...")
df.drop_duplicates(inplace=True, subset="url")

# Remove unnecesary columns
columns = [
    "label",
    "content",
    "title",
    "author",
    "date",
    "last_scraped",
    "categories",
    "url",
]
columns_to_remove = [c for c in df.columns if c not in columns]
print(f"Removing the following columns: {columns_to_remove}")
df.drop(columns_to_remove, inplace=True, axis=1)

print("Cleaned Dataset Shape: ")
print(df)

# Save resulting file
filename = os.path.basename(csv_file_name)
print(f"Saving cleaned data to: {filename}")
with open(filename, "+w") as f:
    df.to_csv(f)



"""
    Use this script to scrape data from primicia with non-relevant information, 
    and then store it in a csv file in the current working directory

    Script Arguments:
        limit (position 1) : max ammount of urls to crawl
"""
# Python imports
import datetime
import sys

# Local imports
from c4v.scraper.crawler.crawlers.primicia_crawler import PrimiciaCrawler
from c4v.scraper.scraper import bulk_scrape

# Third Party imports
import pandas as pd

# Take time:
start = datetime.datetime.now()

# Set up max ammount of urls to crawl:
if len(sys.argv) < 2:
    LIMIT = 10000
    print(f"[WARNING] No Limit provided, defaulting to {LIMIT}")
else:
    try:
        LIMIT = int(sys.argv[1])
    except Exception as e:
        print(
            f"Invalid limit argument. Are you sure it is a valid number?\n\tError: {e}",
            file=sys.stderr,
        )
        exit(1)

# Crawl & scrape urls
print(f"Scraping up to {LIMIT} urls from primicia")

crawler = PrimiciaCrawler.from_irrelevant()
urls = crawler.crawl_urls(up_to=LIMIT)

print(f"URLs scrawled, gathered: {len(urls)}. Scraping new urls...")

data = bulk_scrape(urls)
del urls

print(f"URLs scraped: {len(data)}. Converting to dataframe and saving to csv...")

# Convert to dataframe
df = pd.DataFrame(data)

# Set up datetime suffix
date_suffix = datetime.datetime.strftime(datetime.datetime.now(), "%Y%m%d%H%M%S")
filename = f"primicia_irrelevant_{date_suffix}.csv"

with open(filename, "w+") as file:
    df.to_csv(file)

print(f"Data saved to {filename}")

end = datetime.datetime.now()

print(f"Total Time: {(end - start).total_seconds()}s")



from c4v.data.angostura_loader import AngosturaLoader
import tensorflow as tf
import pandas as pd
import math
import os


class TweetLoader:
    def __init__(self):
        """
        Abstraction of AngosturaLoader to programmatically query Angostura's tweets tables.
        """
        self.angostura_loader = AngosturaLoader()

    def load_sinluz_tweets(
        self, batch_size, batch_number, print_table_info=False, return_tf_dataset=True
    ):
        """
        Queries `angostura.meditweet_sin_luz`.
        Receives batch_size and number of batches to query.
        Yields a generator that returns the queried data.
        ----------------------------------------------------

        Params: 

            batch_size: Int. Default = None.
                Number of tweets to query per batch.
            
            batch_number: Int. Default = None.
                Number of batches to query.

            print_table_info: Bool. Default = False.
                Prints number of records in `angostura.meditweet_sin_luz`.

            return_tf_dataset: Bool. Default = True.
                If True, generator returns a tf.Dataset only containing the column `full_text`.
                If False, generator returns a dataframe with the complete number of columns:
                    id, created_at, full_text, metadata,  user, geo, place

        Example:

            test = TweetLoader()
            test_query = test.load_sinluz_tweets(batch_size=100, batch_number=1)
            
            # Print values
            for x in next(test_query):
                    print(x)

            test_query = test.load_sinluz_tweets(batch_size=100, batch_number=2)
            
            # Loop over queried batches
            for text_batch in test_query:
                print(text_batch)
        """

        # I added this due to a bug -  Your CPU supports instructions that this TensorFlow binary was not compiled to use: AVX AVX2
        # Fix Reference: https://stackoverflow.com/questions/47068709/your-cpu-supports-instructions-that-this-tensorflow-binary-was-not-compiled-to-u
        os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"

        if print_table_info:
            total_tweets = self.angostura_loader.create_query(
                """
                SELECT 
                    count(*) AS count_total
                FROM 
                    `angostura.meditweet_sin_luz` 
                """
            )
            total_records = int(total_tweets.count_total)

            print(
                f"The table `angostura.meditweet_sin_luz` has a total of {total_records} rows."
            )

        for batch in range(batch_number):

            print(f"Loading batch number {batch}")

            query = f"""
                SELECT 
                    *
                FROM 
                    `angostura.meditweet_sin_luz`
                ORDER BY 
                    created_at ASC
                LIMIT {batch_size}
                """

            batch_df = self.angostura_loader.create_query(query)
            target = batch_df.pop("full_text").values

            text_tfdataset = tf.data.Dataset.from_tensor_slices(target)

            if return_tf_dataset:
                yield text_tfdataset
            else:
                yield batch_df



"""
    Format columns of the client dataset in such a way that they match the ScrapedData format, plus a label field:

    url
    last_scraped
    title
    content
    author
    categories
    date
    label
"""
# Python imports
import sys
import os
import datetime
import pytz

# Third party
import pandas as pd
from pandas.core.indexing import is_nested_tuple

# Get command line argument
if len(sys.argv) < 2:
    print("Missing csv filename argument", file=sys.stderr)
    exit(1)

csv_file_name = sys.argv[1]  # name of csv file
with open(csv_file_name) as file:
    df: pd.DataFrame = pd.read_csv(file)

# Rename columns
print("Columnds are: ", df.columns)

df.rename(
    {
        "text": "content",
        "tipo_de_evento": "label",
        "tags": "categories",
        "Link de la noticia ": "url",
        "tipo de evento": "label",
    },
    axis=1,
    inplace=True,
)

# Valid columns
columns = ["url", "title", "content", "author", "categories", "date", "label"]

# Remove irrelevant columns
columns_to_remove = [col for col in df.columns if col not in columns]
print(f"Removing columns: {columns_to_remove}")

df.drop(columns_to_remove, inplace=True, axis=1)
print(f"Columns: {df.columns}")
old_df = df.copy()
url_to_labels_df = old_df.groupby("url")

# Remove duplicates
df.drop_duplicates(inplace=True, ignore_index=True, subset=["url"])

# Try to edit label field
for (url, sub_df) in url_to_labels_df:
    labels = list(set(sub_df.label))
    df.loc[df.url == url, "label"] = [labels]

# Complete missing columns:
df["last_scraped"] = datetime.datetime.now(tz=pytz.utc)

filename = f"cleaned_{os.path.basename(csv_file_name)}"
df.to_csv(filename)



# external imports
import scrapy
from scrapy.crawler import CrawlerProcess
import scrapy.signals

# Project imports
from . import scrapy_settings as settings
from c4v.scraper.scraped_data_classes.scraped_data import ScrapedData

# Python imports
from typing import List


class SpiderManager:
    """
        Utility class to perform common operations in 
        Spider classes
    """

    _process = None

    @classmethod
    def process(cls) -> CrawlerProcess:
        """
            retrieves process object
        """
        if not cls._process:
            cls._process = CrawlerProcess(settings.CRAWLER_SETTINGS)
        return cls._process

    def __init__(self, spider) -> None:

        self.spider = spider
        self._scraped_items = []

        def add_item(item):
            self._scraped_items.append(item)

        self._add_items = add_item

    def parse(self, response) -> ScrapedData:
        """
            return scraped data from a valid response
            Parameters: 
                + response : scrapy.http.Response = response object holding the actual response
            Return:
                dict like object with scraped data
        """
        spider = self.spider()
        return spider.parse(response)

    def scrape(self, url: str) -> ScrapedData:
        """
            Return scraped data from a single Url
            Parameters:
                + url : str = url whose data is to be scraped. Should be compatible with the given spider
            Return:
                dict like object with scraped data
        """
        self.schedule_scraping([url])
        self.start_bulk_scrape()
        scraped = self.get_scraped_items()
        return scraped[0] if scraped else None

    def schedule_scraping(self, urls: List[str]):
        """
            Schedule urls to be scraped
            Parameters:
                + urls : [str] = list of urls to scrape
        """
        # if nothing to do, just return an empty list
        if not urls:
            return

        # set up urls to scrape
        self.spider.start_urls = urls

        # create crawler for this spider, connect signal so we can collect items
        crawler = SpiderManager.process().create_crawler(self.spider)
        crawler.signals.connect(self._add_items, signal=scrapy.signals.item_scraped)

        # start scrapping
        SpiderManager.process().crawl(crawler)

    def start_bulk_scrape(self):
        """
            Scrape stored urls. Note that if multiple spider managers
            called schedule_scraping before, all of them will be scraped,
            not only this one
        """
        if SpiderManager.process().crawlers:
            SpiderManager.process().start()

    def get_scraped_items(self) -> List[ScrapedData]:
        """
            Get items scraped by a scraping process, flushing internal list in the 
            process
            Return:
                List of scraped elements 
        """
        scraped = self._scraped_items
        self._scraped_items = []
        return scraped



"""
    For now, this is going to be the main CLI executable,
    so we can test things in the meanwhile
"""
# Third party imports
from datetime import datetime
import click

# Python imports
from typing import List, Tuple
from urllib.error import HTTPError
import os
from pathlib import Path
import sys

# Local imports
from c4v.scraper.scraped_data_classes.scraped_data import ScrapedData
from c4v.scraper.settings import INSTALLED_CRAWLERS
from c4v.scraper.persistency_manager.sqlite_storage_manager import SqliteManager
from c4v.scraper.utils import data_list_to_table_str
from c4v.scraper.settings import INSTALLED_CRAWLERS
from c4v.config import settings
from c4v.microscope.manager import Manager


# Folder to search for local files
DEFAULT_FILES_FOLDER = settings.c4v_folder
DEFAULT_DB = settings.local_sqlite_db or os.path.join(
    settings.c4v_folder, settings.local_sqlite_db_name
)


@click.group()
def c4v_cli():
    """
        Command entry point 
    """
    # init files if necessary:
    path = Path(DEFAULT_FILES_FOLDER)
    if not path.exists():
        click.echo(f"[INFO] Creating local files folder at: {DEFAULT_FILES_FOLDER}")
        try:
            path.mkdir(parents=True)
        except Exception as e:
            print(f"[ERROR] Could not create '{path}' folder: {e}", err=True)
    elif not path.is_dir():
        click.echo(
            f"[ERROR] Files folder '{path}' already exists but it's not a file.",
            err=True,
        )


@c4v_cli.command()
@click.option(
    "--files",
    is_flag=True,
    help="Interpret url list as files instead of urls, so urls are retrieved from such files. The file is expected to be formatted such that there's one url per line",
)
@click.option(
    "--loud", is_flag=True, help="Print scraped data to stdio as it is being scraped"
)
@click.option(
    "--limit",
    default=-1,
    help="Maximum amount of elements to scrape. Scrape all if no argument given",
)
@click.argument("urls", nargs=-1)
def scrape(
    urls: List[str] = None, files: bool = None, loud: bool = False, limit: int = -1
):
    """
        Use this command to run a scraping process. Scraped urls are retrieved from database by default.\n
        If provided with a list of urls, then those urls are the ones to be scraped and stored. \n
        You can pass the --files flag to interpret those paths as paths to files storing the actual urls,\n
        formated as one url per line.\n
        Parameters:\n
            + urls : [str] = List of urls or files to get urls from\n
            + files : bool = if urls are in files or not, defaults to false\n
            + up_to : int = maximum amount of urls to scrape in this pass. Scrape all non-scraped if not provided\n
            + loud : bool = if should print scraped data once a scraping is finished\n
    """

    manager = Manager.from_default()
    client = CLIClient(manager, urls, files)

    # Read urls
    urls_to_scrape = []

    if not urls:
        urls_to_scrape = [d.url for d in manager.get_all(limit, scraped=False)]
    elif files:  # if urls are stored in files
        urls_to_scrape = client.get_urls(urls)
    else:
        urls_to_scrape = urls

    # scrape urls
    scraped_data = client.get_data_for_urls(urls_to_scrape, should_scrape=True)

    # Print obtained results if requested
    if loud:
        click.echo(data_list_to_table_str(scraped_data))


@c4v_cli.command()
@click.option("--list", is_flag=True, help="list available crawlers")
@click.option("--all", is_flag=True, help="Run all available crawlers")
@click.option("--all-but", is_flag=True, help="Run all crawlers except listed ones")
@click.option("--loud", is_flag=True, help="Print results to terminal")
@click.option("--limit", default=-1, help="Max number of new urls to store")
@click.argument("crawlers", nargs=-1)
def crawl(
    crawlers: List[str] = [],
    list: bool = False,
    all: bool = False,
    all_but: bool = False,
    loud: bool = False,
    limit: int = -1,
):
    """
        Crawl for new urls, ignoring already scraped ones.\n
        Parameters:\n
            + list : bool = if should print list of available crawlers by name\n
            + all  : bool = if should crawl all crawlers. Not true by default as a crawling process may be really slow\n
            + all_but : bool = run all crawlers except the ones provided as arguments\n
            + loud : bool = if should print scraped urls to stdio\n
    """

    # Create default CLI client
    client = CLIClient()

    # list available crawlers if requested to
    crawlable_sites = "".join([f"\t{crawl.name}\n" for crawl in INSTALLED_CRAWLERS])
    if list:
        click.echo(f"Crawlable sites: \n{crawlable_sites}")
        if not crawlers and not all:  # if nothing else to do, just end
            return

    # Check for errors:
    # if no crawlers to use, just end
    if not crawlers and not all:
        click.echo(f"Not crawlers provided. Available crawlers:\n{crawlable_sites}")
        return

    # raise a warnning if incompatible flags where provided
    if all_but and all:
        click.echo(
            f"[WARNING] --all and --all-but incompatible flags were provided, using only --all"
        )

    # set up crawlers to run
    if all:
        crawlers_to_run = [crawler.name for crawler in INSTALLED_CRAWLERS]
    elif all_but:
        crawlers_to_run = [
            crawler.name
            for crawler in INSTALLED_CRAWLERS
            if crawler.name not in crawlers
        ]
    else:
        crawlers_to_run = [
            crawler.name for crawler in INSTALLED_CRAWLERS if crawler.name in crawlers
        ]

    client.crawl_new_urls_for(crawlers_to_run, limit, loud)


@c4v_cli.command()
@click.option("--urls", is_flag=True, help="Only list urls")
@click.option("--limit", default=100, help='List only up to "limit" rows')
@click.option("--col-len", default=50, help="Columns max length")
@click.option("--count", is_flag=True, help="Print only count of selected data")
@click.option(
    "--scraped-only",
    default=None,
    help="Retrieve only complete rows, those with its scraped data",
)
def list(
    urls: bool = False,
    limit: int = 100,
    col_len: int = 50,
    count: bool = False,
    scraped_only: bool = None,
):
    """
    List requested info as specified by arguments.\n
    Parameters:\n
        + urls : bool = list urls only\n
        + limit : int = max num of rows to show. Set -1 to list all\n
        + col_len : int = Maximum lenght of shown columns\n
        + scraped_only : bool = if true, retrieve already scraped ones only, if false, retrieve non scraped only. If not provided, return anything\n
    """

    scraped_only = (
        (scraped_only == "true" or scraped_only == "True" or scraped_only == "1")
        if scraped_only
        else None
    )

    db_manager = SqliteManager(DEFAULT_DB)

    # Just print urls if requested so
    if urls:
        for data in db_manager.get_all(limit):
            click.echo(data.url)
        return

    data = [d for d in db_manager.get_all(limit, scraped_only)]

    if count:
        click.echo(len(data))
        return

    # Get printable version of retrieved data
    data_to_print = data_list_to_table_str(data, max_cell_len=col_len)

    click.echo(data_to_print)


@c4v_cli.command()
@click.option(
    "--no-scrape", is_flag=True, help="Don't scrape if url is not found in DB"
)
@click.option("--file", is_flag=True, help="Get urls of news to classify from a file")
@click.option(
    "--limit",
    is_flag=False, 
    default=-1,
    help="Limit how much instances to classify in this run. Specially usefull when classifying pending data, if less than 0, then select as much as you can (default). Otherwise, classify at the most the given number",
    type=int,
)
@click.argument("inputs", nargs=-1)
def classify(
    inputs: List[str] = [], no_scrape: bool = False, file: bool = False, limit: int = -1
):
    """
        Run a classification over a given url or from a file, using the model stored in the provided
        experiment. Usage:
            c4v classify <branch_name>/<experiment_name> <url>
    """
    # Validate input:
    n_args = len(inputs)
    if (
        n_args < 2
    ):  # Get at the least 2 args, experiment as branch/experiment and some url
        click.echo(
            "[ERROR] Should provide at the least 2 arguments, experiment and at the least 1 url"
        )
        return

    # Create manager object
    manager = Manager.from_default()
    client = CLIClient(manager, file)

    # validate branch and name
    parsed_branch_and_name = CLIClient.parse_branch_and_experiment_from(inputs[0])
    if parsed_branch_and_name == None:
        return
    else:
        branch, experiment = parsed_branch_and_name

    # check if we have to classify pending data
    classify_pending = n_args == 2 and inputs[1] == "pending"
    if classify_pending:
        res = manager.run_pending_classification_from_experiment(
            branch, experiment, save=True, limit=limit
        )
        click.echo(f"[INFO] {len(res)} classified rows")
        return

    data = client.get_data_for_urls(urls=inputs[1:], should_scrape=not no_scrape)

    # Do nothing if not necessary:
    if not data:
        click.echo("[INFO] Nothing to classify")
        return

    # Try to classify given data
    try:
        results = manager.run_classification_from_experiment(branch, experiment, data)
    except ValueError as e:
        click.echo(f"[ERROR] Could not classify provided data.\n\tError: {e}", err=True)
        return
    except ModuleNotFoundError as e:
        click.echo(f"[ERROR] Could not found some modules, maybe you should try to change the installation profile of c4v. Erro: {e}", err=True)
        return

    # Pretty print results:
    for result in results:
        click.echo("\n")
        data: ScrapedData = result["data"]
        scores = result["scores"]
        click.echo(f"\t{data.title if data.title else '<no title>'} ({data.url})")
        click.echo(f"\t\t{data.label}")
        click.echo(f"\t\t{scores}")


@c4v_cli.command()
@click.option(
    "--no-scrape", is_flag=True, help="Don't scrape if data is not available locally"
)
@click.argument("url", nargs=1)
def show(url: str, no_scrape: bool = False):
    """
        Show the entire data for a given URL
    """
    # Create manager object
    manager = Manager.from_default()
    client = CLIClient(manager, [url])

    data = client.get_data_for_urls(should_scrape=not no_scrape)

    # Check if could retrieve desired element
    if not data:
        click.echo("[EMPTY]")
        return
    else:
        element = data[0]

    # Pretty print element:
    line_len = os.get_terminal_size().columns
    click.echo("+" + ("=" * (line_len - 2)) + "+")
    click.echo(f"\tUrl        : {element.url}")
    click.echo(f"\tTitle      : {element.title}")
    click.echo(f"\tAuthor     : {element.author}")
    click.echo(f"\tDate       : {element.date}")
    click.echo(
        f"\tCategories : {', '.join(element.categories) if element.categories else '<No Category>'}"
    )
    click.echo(
        f"\tScraped    : {datetime.strftime(element.last_scraped, settings.date_format)}"
    )
    click.echo("=" * line_len)
    cleaned_content = "\n".join(
        [s for s in [s.strip() for s in element.content.splitlines()] if s != ""]
    )
    click.echo(cleaned_content)
    click.echo("+" + ("=" * (line_len - 2)) + "+")


@c4v_cli.command()
@click.option("--url", is_flag=True, help="Interpret string as URL")
@click.option(
    "--no-scrape",
    is_flag=True,
    help="In case of retrieving data from URL, don't scrape it if not available",
)
@click.option(
    "--html",
    default="./explanation.html",
    help="Dump results in an human readable format to an html file",
    nargs=1,
)
@click.option(
    "--label",
    default=None,
    help="Label to include in expalantion. If the predicted label is different from this one, then explain how much this label was contributing to its corresponding value",
)
@click.argument("experiment", nargs=1)
@click.argument("sentence", nargs=1)
def explain(
    experiment: str,
    sentence: str,
    url: bool = False,
    no_scrape: bool = False,
    html: str = None,
    label: str = None,
):
    """
        Show explainability for the given string. That is, show how much each word contributes 
        to each tag in the classifier. The result depends on the experiment, as it will load that 
        model to explain. The experiment argument follows the <branch_name>/<experiment_name> syntax
        Arguments:
            experiment : str = experiment format, following <branch_name>/<experiment_name> format
            sentence   : str = expression to explain 
    """
    microscope_manager = Manager.from_default()
    client = CLIClient(microscope_manager)

    # Get text to explain
    if url:
        # Get data to classify
        datas = client.get_data_for_urls([sentence], not no_scrape)

        # check if there's data retrieved to explain
        if not datas:
            click.echo("Nothing to explain")
            return

        # Get content from data retrieved
        text_to_explain = datas[0].content
    else:
        text_to_explain = sentence

    # Parse branch name and experiment
    branch_and_experiment = client.parse_branch_and_experiment_from(experiment)
    if branch_and_experiment == None:
        return

    # unpack branch name and experiment name
    branch, experiment = branch_and_experiment

    # Check label input
    try:
        possible_labels = microscope_manager.get_classifier_labels()
    except ModuleNotFoundError as e:
        click.echo(f"[ERROR] Could not found some modules, maybe you should try to change the installation profile of c4v. Erro: {e}", err=True)
        return

    if label and label not in possible_labels:
        click.echo(
            f"[WARNING] Provided label not a valid label, ignoring label argument {label}.",
            err=True,
        )
        click.echo(f"Possible Labels:", err=True)
        for l in possible_labels:
            click.echo(f"\t* {l}", err=True)
        label = None

    # try to explain
    try:
        explanation = microscope_manager.explain_for_experiment(
            branch, experiment, text_to_explain, html_file=html, additional_label=label
        )
    except ValueError as e:
        click.echo(f"[ERROR] Could not explain given sentence. Error: {e}")
        return

    # Pretty print results
    scores = explanation["scores"]
    label = explanation["label"]
    click.echo(f"Predicted Label: {label}\nScores:")
    for (word, score) in scores:
        click.echo(f"\t* {word} : {score}")


@c4v_cli.group()
def experiment():
    """
        Experiment Management. You can get info about experiments with this command, such as 
        listing, and removing them if no longer necessary
    """
    path = Path(settings.experiments_dir)
    if not path.exists():
        click.echo(f"[INFO] Creating experiments folder in: {path}")
        try:
            path.mkdir()
        except Exception as e:
            click.echo(
                f"[ERROR] Could not create folder due to the following error: {e}",
                err=True,
            )

    elif not path.is_dir():
        click.echo(
            f"[ERROR] Could not create folder {path}. File already exists but is not a folder",
            err=True,
        )


@experiment.command()
@click.argument("branch", nargs=1, required=False)
def ls(branch: str = None):
    """
        List branches if no argument is provided. If a branch name is specified, then list experiments within that branch.
        Examples:
            c4v experiment ls
                branch1
                branch2
                branch3
            c4v experiment ls branch1
                experiment1
                experiment2
    """
    # TODO tal vez mover esta lógica al ExperimentFSManager?
    if not branch:
        click.echo(f"[INFO] Listing from {settings.experiments_dir}")
        files = CLIClient._ls_files(settings.experiments_dir)
        click.echo("\n".join(files))
        return

    # Check if branch exists
    path = Path(settings.experiments_dir, branch)
    if not path.exists():
        click.echo(
            f"[ERROR] This is not a valid branch name: {branch} in {path}. You can see available branches using the command:\n\tc4v experiment ls",
            err=True,
        )
        return
    elif not path.is_dir():
        click.echo(
            f"[ERROR] Invalid Branch path: {path}. The branch name '{branch}' does not refers to an actual branch's directory"
        )

    # As everything is ok, just list files
    click.echo(f"[INFO] Listing from {path}")
    files = CLIClient._ls_files(path=str(path))
    click.echo("\n".join(files))


@experiment.command()
@click.argument("experiment", nargs=1)
def summary(experiment: str):
    """
        Print summary for an existent experiment given its name with branch and experiment name\n
        Example:\n
            `c4v experiment summary branch_name/experiment_name`
    """

    # Parse branch and experiment name
    branch_and_experiment = CLIClient.parse_branch_and_experiment_from(experiment)
    if not branch_and_experiment:
        return

    # as everything went ok, parse branch name and experimet
    branch_name, experiment_name = branch_and_experiment

    path = Path(settings.experiments_dir, branch_name, experiment_name, "summary.txt")
    # Check if file exists
    if not path.exists():
        click.echo(
            f"[ERROR] Summary for experiment {experiment} not found in {path}", err=True
        )
        return
    elif not path.is_file():
        click.echo(
            f"[ERROR] Sumamry for experiment {experiment} in {path} is not a valid file",
            err=True,
        )
        return

    # As everything went ok, print file content
    click.echo(f"[INFO] Reading summary from: {path}")
    click.echo(path.read_text())


class CLIClient:
    """
        This class will manage common operations performed by the CLI tool
    """

    def __init__(
        self, manager: Manager = None, urls: List[str] = [], from_files: bool = False
    ):

        # Default manager
        if not manager:
            manager = Manager.from_default(local_files_path=settings.c4v_folder)

        self._manager = manager

        # If urls are in files, load them from such files
        if from_files:
            self._urls = CLIClient._parse_lines_from_files(urls)
        else:
            self._urls = urls

    def get_data_for_urls(
        self, urls: List[str] = None, should_scrape: bool = True
    ) -> List[ScrapedData]:
        """
            Return a list of ScrapedData from a list of urls.
            Parameters:
                urls : [str] = List of urls whose data is requested. If not provided, defaults to the 
                                list stored whithin this class
                should_scrape : bool = If should scrape data for urls that are not currently available
            Return:
                List of retrieved urls
        """
        urls_to_retrieve = urls or self._urls

        # Check scrapable urls:

        scrapable_urls, non_scrapables = self._manager.split_non_scrapable(
            urls_to_retrieve
        )

        # Warn the user that some urls won't be scraped
        if non_scrapables:
            click.echo(
                "[WARNING] some urls won't be retrieved, as they are not scrapable for now.",
                err=True,
            )
            click.echo("Non-scrapable urls:", err=True)
            click.echo("\n".join([f"\t* {url}" for url in non_scrapables]), err=True)

        # check if some http error happens
        data = []
        try:
            data = self._manager.get_bulk_data_for(
                scrapable_urls, should_scrape=should_scrape
            )
        except HTTPError as e:
            click.echo(
                f"[ERROR] Could not scrape all data due to connection errors: {e}",
                err=True,
            )

        # Tell the user if some urls where not retrieved
        succesfully_retrieved = {d.url for d in data}
        if any(url not in succesfully_retrieved for url in scrapable_urls):
            click.echo(f"[WARNING] Some urls couldn't be retrieved: ")
            click.echo(
                "\n".join(
                    [
                        f"\t* {url}"
                        for url in scrapable_urls
                        if url not in succesfully_retrieved
                    ]
                )
            )

        return data

    def get_urls(self) -> List[str]:
        """
            Return stored urls
        """
        return self._urls

    @staticmethod
    def parse_branch_and_experiment_from(line: str) -> Tuple[str, str]:
        """
            Return a tuple with branch name and experiment name from a line representing 
            and experiment name. Report error if not a valid name
            Parameters:
                line : str = line to parse as branch name and experiment name. For example:
                        branch/exp1
                        branch-exp1
                    Invalid examples:
                        branch exp1 
                        branch\exp1
            Return:
                A tuple, with the first member as the branch name ans the second one as the experiment name
        """
        separators = ["/", "-"]
        for separator in separators:
            branch_and_name = line.split(separator)
            if len(branch_and_name) == 2:
                branch, name = branch_and_name
                return (branch, name)

        click.echo(
            f"[ERROR] Given experiment name is not valid: {line}. Should be of the form:",
            err=True,
        )
        for separator in separators:
            click.echo(f"\tbranch_name{separator}experiment_name")
        return None

    def crawl_new_urls_for(
        self, crawler_names: List[str], limit: int = -1, loud: bool = False
    ):
        """
            Crawl URLs for the given crawlers, up to a max number of urls.
            Parameters:
                crawlers : [str] = List of crawlers names to use
                limit    : int = Max limit of urls to get. 
                loud     : bool = If should print to terminal obtained scrapers. False by default
        """
        # function to format crawled urls list
        format_url_list = lambda list: "".join([f"{s}\n" for s in list])

        def process(list: List[str]):
            if loud:
                click.echo(format_url_list(list))

        # Names for installed crawlers
        crawlers = [c.name for c in INSTALLED_CRAWLERS]

        not_registered = [name for name in crawler_names if name not in crawlers]

        # Report warning if there's some non registered crawlers
        if not_registered:
            click.echo(
                "WARNING: some names in given name list don't correspond to any registered crawler.",
                err=True,
            )
            click.echo(
                "Unregistered crawler names: \n"
                + "\n".join([f"\t* {name}" for name in not_registered]),
                err=True,
            )

        self._manager.crawl_and_process_new_urls_for(
            [c for c in crawler_names if c not in not_registered], process, limit=limit
        )

    @staticmethod
    def _parse_lines_from_files(files: List[str]) -> List[str]:
        """
            Utility function to collect all lines from multiple files
        """
        lines = []
        for file in files:  # iterate over every file
            try:
                with open(file) as file_with_lines:
                    content = map(
                        lambda s: s.strip(), file_with_lines.readlines()
                    )  # parse every line as a single url
                    lines.extend(content)
            except IOError as e:
                click.echo(
                    f"Could not open input file: {file}. Error: {e.strerror}", err=True
                )
        return lines

    @staticmethod
    def _ls_files(path: str) -> List[str]:
        """
            Returns a list of file names within a given directory 
        """
        return [str(x.name) for x in Path(path).glob("*")]


if __name__ == "__main__":
    c4v_cli()



"""
    Misc. helper functions.
"""
# External imports
from bs4 import BeautifulSoup
from tabulate import tabulate

# local imports
from c4v.scraper.scraped_data_classes.scraped_data import ScrapedData
from c4v.scraper.scrapers.base_scraper import BaseScraper

# Python imports
import re
import pytz
from datetime import datetime
from urllib.parse import urlparse
from typing import List, Type


def strip_http_tags(element: str) -> str:
    """
        Get text from a html formated string
        Parameters:
            + element - str : html formated string to be cleaned
        Return:
            String cleaned from html tags, leaving text only
    """

    # use BeautifulSoup to clean string
    soup = BeautifulSoup(element, "lxml")
    return soup.get_text()


def get_element_text(selector: str, response) -> str:
    """
        Return cleaned text from an element selected by "selector".
        May return None if element was not found
        Parameters: 
            + selector - str : valid css selector used to select html element 
            + response : response as passed to parse in Spider class
        Return:
            Selected element's content as a cleaned string, or None if such 
            element is not present
    """
    # try to select element from document
    value = response.css(selector)

    # in case nothing found, return None
    if not value:
        return None

    # clean from html tags
    soup = BeautifulSoup(value.get(), "lxml")

    return soup.get_text()


def valid_url(url: str) -> bool:
    """
        Check that the given url is actually a valid url
    """
    url_matcher = re.compile(
        r"^(?:http|ftp)s?://"  # http:// or https://
        r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"  # domain...
        r"localhost|"  # localhost...
        r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # ...or ip
        r"(?::\d+)?"  # optional port
        r"(?:/?|[/?]\S+)$",
        re.IGNORECASE,
    )

    return re.match(url_matcher, url) is not None


def get_domain_from_url(url: str) -> str:
    """
        Get domain name from  a valid url 
        Parameters:
            + url : str = url to get domain from
        Return:
            domain name for the given url
    """

    return urlparse(url).netloc


def check_scrapers_consistency(scrapers: List[Type[BaseScraper]]):
    """
        Check consistency for a scraper list
    """
    for scraper in scrapers:
        check_scraper_consistency(scraper)


def check_scraper_consistency(scraper: Type[BaseScraper]):
    """
        Check consistency for a scraper. This function checks that:
            + Scraper provides intended domain
    """
    assert scraper.intended_domain != None and isinstance(
        scraper.intended_domain, str
    ), f"Scraper {scraper} does not provide intended_domain"


def get_datetime_now() -> datetime:
    """
        Return datetime formated properly
    """
    return datetime.now(tz=pytz.UTC)


def data_list_to_table_str(data_list: List[ScrapedData], max_cell_len: int = 50) -> str:
    """
        Return a string representation of a given data list in a tabular format
        Parameters:
            + data_list : [ScrapedData] = list of data instances to list in table
    """

    cols = [k for k in ScrapedData.__dataclass_fields__.keys()]
    cols.sort()
    headers = ["row"] + cols

    truncate = lambda s: s if len(s) <= max_cell_len else s[:max_cell_len] + "..."

    rows = [
        [i + 1] + [truncate(str(data.__getattribute__(attr))) for attr in cols]
        for (i, data) in enumerate(data_list)
    ]

    return tabulate(rows, headers=headers)



# Internal imports
from .. import utils
from c4v.scraper.scraped_data_classes.elpitazo_scraped_data import ElPitazoData

# External imports
import scrapy
from scrapy.http import Response

# Python imports
from typing import List, Dict, Any
from datetime import datetime
import pytz


class ElPitazoSpider(scrapy.Spider):
    """
        Spider to scrape ElPitazo data 
    """

    name = "el_pitazo"

    start_urls = []

    def parse(self, response: Response) -> ElPitazoData:
        """
            Returns a data object describing data scrapable for this 
            scraper

            Parameters:
                + response : Response = scrapy response object
            
            Return:
                An ElPitazoData instance 
        """

        # These are simple properties, just get its text with a valid
        # selector
        title = response.css(".tdb-title-text ::text").get() or ""
        date = response.css(".entry-date ::text").get() or ""
        author = response.css(".tdb-author-name ::text").get() or ""

        body = self._get_body(response)

        tags = self._get_tags(response)

        # categories
        categories = response.css(".tdb-entry-category ::text").getall()

        return ElPitazoData(
            body=body,
            tags=tags,
            categories=categories,
            title=title,
            author=author,
            date=date,
            url=response.url,
            last_scraped=utils.get_datetime_now(),
        )

    def _get_body(self, response: Response) -> str:
        """
            Get article body as a single string
        """
        body = response.css("#bsf_rt_marker > p").getall()
        body = filter(lambda p: p.startswith("<p>") and p.endswith("</p>"), body)
        body = map(utils.strip_http_tags, body)

        body = "\n".join(body)

        return body.strip()

    def _get_tags(self, response: Response) -> List[str]:
        """
            Try to get tags from document if available
        """
        tags = response.css(".tdb-tags > li > a ::text").getall()
        return tags
